#==============================================================================

# Copyright (c) 2016-11-21, Laurenz Wiskott, http://www.ini.rub.de/PEOPLE/wiskott/
# License: BSD 2-Clause, https://opensource.org/licenses/BSD-2-Clause, see the end

#==============================================================================

import numpy as np

from gaussian import Gaussian
from gaussian_cluster import GaussianCluster

class Mixture:
    
    COLORS = ['r','b','g','m','c','y','k']
    NUM_COMPONENTS_DEFAULT = 3
    COMPONENT_CLASS = None

    def __init__(self, dim=Gaussian.DIM_DEFAULT, num_components=NUM_COMPONENTS_DEFAULT,
                 mean_std=Gaussian.MEAN_STD_DEFAULT, num=GaussianCluster.NUM_DEFAULT,
                 target_points=None):

        self.dim = dim
        self.num_components = num_components
        self.num = num * np.ones(self.num_components, dtype=int)
        self.prob_components = np.arange(self.num_components)
        for c_idx in range(self.num_components):
            self.prob_components[c_idx] = 1./self.num_components
        self.mean_std = mean_std
        self.components = list(range(self.num_components))
        if self.COMPONENT_CLASS is not None:
            for c_idx in range(self.num_components):
                self.components[c_idx] = self.COMPONENT_CLASS(dim=self.dim, mean_std = self.mean_std, 
                                                            num=self.num[c_idx], color=self.COLORS[c_idx%len(self.COLORS)])
        self._prob_component = np.array([1./self.num_components for c_idx in range (self.num_components)])
        if target_points is not None:
            self.target_points = target_points
        elif self.COMPONENT_CLASS == GaussianCluster:
            self.target_points = self.points()
        else:
            self.target_points = None
            print("Initialization of points for MixtureOfGaussians not yet implemented!")
            exit()
        if self.COMPONENT_CLASS == Gaussian and self.target_points is not None:
            for c_idx in range(self.num_components):
                self.components[c_idx].mean = self.target_points[np.random.random_integers(0,len(self.target_points)-1)]
        self._mean = np.array([self.components[c_idx].mean for c_idx in range(self.num_components)])
        self._cov = np.array([self.components[c_idx]._cov for c_idx in range(self.num_components)])


    def ax_plot(self, ax, axes=(0,1), show_target=True):

        if self.target_points is not None and show_target:
            ax.scatter(self.target_points[:,axes[0]], self.target_points[:,axes[1]], s=self.COMPONENT_CLASS.POINT_SIZE, color='k')
        for c_idx in range(self.num_components):
            self.components[c_idx].ax_plot(ax, axes=(0,1), linewidth=self._prob_component[c_idx]*Gaussian.LINEWIDTH_DEFAULT)


    def prob_component(self, component_idx):

        return 1./self.num_components


    def prob_components(self):

        return [1./self.num_components for component_idx in range(self.num_components)]


    def pd_points_given_component(self, points, component_idx):

        points = np.array(points)

        try:
            return self.components[component_idx].pd_points(points)
        except np.linalg.LinAlgError:
            print("Exception in 'Mixture.pd_points_given_component': Covariance matrix is probably singular!")
            raise


    def pd_points(self, points):

        points = np.array(points)

        pds_sum = 0
        try:
            for c_idx in range(self.num_components):
                pds_sum += self.prob_component(c_idx) * self.pd_points_given_component(points, c_idx)
        except np.linalg.LinAlgError:
            print("Exception in 'Mixture.pd_points': Covariance matrix is probably singular!")
            raise

        return pds_sum


    def prob_component_given_points(self, component_idx, points):

        points = np.array(points)

        return np.average(self.pd_points_given_component(points, component_idx) \
                          * self.prob_component(component_idx) / self.pd_points(points))


    def update_pd_point_given_component(self, points):
        """Update _pd_point_given_component using function pd_points given mean and cov.

        p(x_n|l) = <_pd_point_given_component> has shape (num_points, num_components).
        """

        self._pd_point_given_component = []
        for c_idx in range(self.num_components):
            try: 
                self._pd_point_given_component.append(self.components[c_idx].pd_points(points))
            except np.linalg.LinAlgError:
                print("Exception in 'Mixture.update_pd_point_given_component': Covariance matrix is probably singular!")
                raise
        self._pd_point_given_component = np.array(self._pd_point_given_component).transpose() # (pnt,cmp)


    def update_pd_point(self, points):
        """Update _pd_point given _pd_point_given_component and _prob_component.

        The equation follows simple probability theory:

            p(x_n) = Sum_l p(x_n|l) P(l)  

        p(x_n)   = <_pd_point>                   has shape (num_points).
        p(x_n|l) = <_pd_point_given_component>   has shape (num_points, num_components).
        P(l)     = <_prob_component>             has shape (            num_components).
        """

        # EXERCISE:
        # self._pd_point = ... # np.array with shape (pnt)


    def update_prob_component_given_point(self, points):
        """Update prob_component_given_point given _pd_point_given_component and _prob_component.

        The equation follows the Bayesian theorem:

                        p(x_n|k) P(k)
            P(k|x_n) = ---------------
                        p(x_n)

        P(k|x_n) = <prob_component_given_point>  has shape (            num_components, num_points).
        p(x_n|k) = <_pd_point_given_component>   has shape (num_points, num_components).
        P(k)     = <_prob_component>             has shape (            num_components).
        """

        # EXERCISE:
        # self.prob_component_given_point = ... # np.array with shape (cmp,pnt)


    def update_mean(self, points):
        """Update mean given prob_component_given_point and point.

        The equation can be found in my lecture notes on clustering in the section "Unisotropic Gaussian":

                   Sum_n P(k|x_n) x_n
            c_k = ---------------------
                   Sum_m P(k|x_m)

        c_k      = <mean>                       has shape (num_components,             num_dimensions).
        P(k|x_n) = <prob_component_given_point> has shape (num_components, num_points).
        x_n      = <point>                      has shape (                num_points, num_dimensions).
        P(k|x_n) x_n                            has shape (num_components, num_points, num_dimensions).
        """

        # EXERCISE:
        # self._mean = ... # np.array with shape (cmp,dim)

        for c_idx in range(self.num_components):
            self.components[c_idx].mean = self._mean[c_idx]


    def update_cov(self, points):
        """Update _cov given prob_component_given_point, point, and mean.

        The equation can be found in my lecture notes on clustering in the section "Unisotropic Gaussian":

                       Sum_n P(k|x_n) (x_n-c_k)(x_n-c_k)^T
            Sigma_k = -------------------------------------
                       Sum_m P(k|x_m)

        Sigma_k  = <_cov>                       has shape (num_components,             num_dimensions, num_dimensions).
        P(k|x_n) = <prob_component_given_point> has shape (num_components, num_points).
        x_n      = <point>                      has shape (                num_points, num_dimensions).
        P(k|x_n) (x_n-c_k)(x_n-c_k)^T           has shape (num_components, num_points, num_dimensions, num_dimensions).
        """

        cmp = len(self.components)
        pnt = len(points)
        dim = len(points[0])

        # EXERCISE:
        # ...
        # self._cov = ... # np.array with shape (cmp,dim,dim)

        for c_idx in range(self.num_components):
            try:
                self.components[c_idx].cov(self._cov[c_idx])
            except np.linalg.LinAlgError:
                print("Exception in 'Mixture.update_cov': Covariance matrix is probably singular!")
                raise


    def update_prob_component(self, points):
        """Update _prob_component given prob_component_given_point.

        The equation can be found in my lecture notes on clustering in the section "Unisotropic Gaussian":

            P(k) = Sum_n P(k|x_n) / N
        
        P(k)     = <_prob_component>            has shape (num_components).
        P(k|x_n) = <prob_component_given_point> has shape (num_components, num_points).
        N        = <num_points>                 has shape (1).
        """

        pnt = len(points)
        # EXERCISE:
        # self._prob_component = ... # np.array with shape (cmp)


    def iterate_fit(self, points):

        mean_old = self._mean
        cov_old = self._cov
        try:
            self.update_pd_point_given_component(points)
            self.update_pd_point(points)
            self.update_prob_component_given_point(points)
            self.update_mean(points)
            self.update_cov(points)
            self.update_prob_component(points)
        except np.linalg.LinAlgError:
            print("Exception in 'Mixture.iterate_fit': Covariance matrix is probably singular!")
            raise

        # Return total squared changes of means and covariance matrices as measures of convergence.
        return np.sum((mean_old - self._mean)**2), np.sum((cov_old - self._cov)**2)


#==============================================================================
#==============================================================================

# BSD 2-Clause License

# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#==============================================================================

